@@ -7,14 +7,20 @@ from ..util import properties
 
 try:
     import dbus
+    from dbus.mainloop.glib import DBusGMainLoop
 except ImportError:
     pass
 
 
 class DBusKeyring(KeyringBackend):
-    """KDE KWallet via D-Bus"""
+    """
+    KDE KWallet 5 via D-Bus
+    """
 
     appid = 'Python program'
+    wallet = None
+    bus_name = 'org.kde.kwalletd5'
+    object_path = '/modules/kwalletd5'
 
     @properties.ClassProperty
     @classmethod
@@ -22,13 +28,15 @@ class DBusKeyring(KeyringBackend):
         if 'dbus' not in globals():
             raise RuntimeError('python-dbus not installed')
         try:
-            bus = dbus.SessionBus()
+            bus = dbus.SessionBus(mainloop=DBusGMainLoop())
         except dbus.DBusException as exc:
             raise RuntimeError(exc.get_dbus_message())
         try:
-            bus.get_object('org.kde.kwalletd5', '/modules/kwalletd5')
+            bus.get_object(cls.bus_name, cls.object_path)
         except dbus.DBusException:
-            raise RuntimeError('cannot connect to org.kde.kwalletd5')
+            tmpl = 'cannot connect to {bus_name}'
+            msg = tmpl.format(bus_name=cls.bus_name)
+            raise RuntimeError(msg)
         return 4.9
 
     def __init__(self, *arg, **kw):
@@ -60,10 +68,10 @@ class DBusKeyring(KeyringBackend):
     def connected(self, service):
         if self.handle >= 0:
             return True
-        bus = dbus.SessionBus()
+        bus = dbus.SessionBus(mainloop=DBusGMainLoop())
         wId = 0
         try:
-            remote_obj = bus.get_object('org.kde.kwalletd5', '/modules/kwalletd5')
+            remote_obj = bus.get_object(self.bus_name, self.object_path)
             self.iface = dbus.Interface(remote_obj, 'org.kde.KWallet')
             self.handle = self.iface.open(
                         self.iface.networkWallet(), wId, self.appid)
@@ -82,8 +90,9 @@ class DBusKeyring(KeyringBackend):
             return None
         if not self.iface.hasEntry(self.handle, service, username, self.appid):
             return None
-        return self.iface.readPassword(
+        password = self.iface.readPassword(
             self.handle, service, username, self.appid)
+        return str(password)
 
     def set_password(self, service, username, password):
         """Set password for the username of the service
@@ -103,3 +112,17 @@ class DBusKeyring(KeyringBackend):
         if not self.iface.hasEntry(self.handle, service, username, self.appid):
             raise PasswordDeleteError("Password not found")
         self.iface.removeEntry(self.handle, service, username, self.appid)
+
+
+class DBusKeyringKWallet4(DBusKeyring):
+    """
+    KDE KWallet 4 via D-Bus
+    """
+
+    bus_name = 'org.kde.kwalletd'
+    object_path = '/modules/kwalletd'
+
+    @properties.ClassProperty
+    @classmethod
+    def priority(cls):
+        return super(DBusKeyringKWallet4, cls).priority - 1
